#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#ifdef linux
#include <dirent.h>
#else
#include <sys/dirent.h>
#endif
#include <sys/stat.h>

int allfiles = 0;	/* Include .xxx files */
int samefiles = 0;	/* print `same:' entries */
int comp_mtime=0;	/* Compare mtimes instead of contents */
int touchsame=0;	/* If contents are same, make times same */
int flushline=0;	/* make printf's line-buffered */
int bothdifferfile=0;	/* print both filenames on differ: lines */
int bothonlyfile=0;	/* print both filenames on only: lines */
int bothsamefile=0;	/* print both filenames on same: lines */

char *prog;

extern char *error_prog;

#define USAGE (void) fprintf(stderr, \
    "Usage: %s [-ahlmt21] [ -X excluded_pats ] dir1  dir2\n", prog)

char **excludepat;
int nexclude=0;

main(argc, argv)
int argc;
char **argv;
{
    /* Compares two directory trees.
     * Symbolic links are followed; if one file is linked to the corresponding
     * file in the other directory, they will compare the same.

     */

    void getoptions();
    int i;
    int compdir();
    char f1[MAXPATHLEN], f2[MAXPATHLEN];

    error_prog = prog = argv[0];

    getoptions(argc, argv, f1, f2);


    /* Strip trailing `/' from names */
    i = strlen(f1);
    if (f1[i-1] == '/')
	f1[i-1] = '\0';
    i = strlen(f2);
    if (f2[i-1] == '/')
	f2[i-1] = '\0';

    if (compdir(f1, f2) == -1)
	exit(1);
    else
	exit(0);
}

/****************************************************************************/
void getoptions(argc, argv, f1, f2)
int argc;
char **argv;
char *f1, *f2;
{
    int i;
    char *s;

    argv++;
    argc--;
    while (*argv && **argv == '-') {
	for (s = *argv+1; *s; s++) {
	    switch (*s) {
	    case 'a':
		allfiles = 1;
		break;
	    case 'h':
		USAGE;
		(void) fprintf(stderr,
		    "-a: compare all files, including `.xxx' files\n");
		(void) fprintf(stderr,
		    "-h: this message\n");
		(void) fprintf(stderr,
		    "-l: make output line buffered\n");
		(void) fprintf(stderr,
		    "-m: compare mtimes instead of contents\n");
		(void) fprintf(stderr,
		    "-s: print `same:' lines when files match\n");
		(void) fprintf(stderr,
"-t: if contents are same, touch mtime on newer file to match older file\n");
		(void) fprintf(stderr,
"-1: print both filenames for `differ:' & `same:' lines; one for `only:' lines\n");
		(void) fprintf(stderr,
"-2: print both filenames on `differ:', `same:', and `only:' output lines\n");
		(void) fprintf(stderr,
"-X pats ... : files matching these sh(1) wildcard patterns aren't compared\n");
		(void) fprintf(stderr,
"    If pat has no slash, only the last component of each name is compared;\n\t\
otherwise the entire name is compared\n");
		exit(0);
		break;
	    case 'l':
		/* setlinebuf(stdout); */
		setvbuf(stdout, (char *) 0, _IOLBF, 0);
		break;
	    case 'm':
		comp_mtime = 1;
		break;
	    case 's':
		samefiles = 1;
		break;
	    case 't':
		touchsame = 1;
		break;
	    case '1':
		bothdifferfile = 1;
		bothsamefile = 1;
		bothonlyfile = 0;
		break;
	    case '2':
		bothdifferfile = 1;
		bothsamefile = 1;
		bothonlyfile = 1;
		break;
	    case 'X':
		excludepat = argv + 1;
		for (argv++, argc--;
		    argc > 2 && **argv != '-' ; argv++, argc--) {
		    nexclude++;
		}
		/* Allow for fact that arg[cv] are --/++'d again at bottom */
		argv--; argc++;
		break;
	    default:
		USAGE;
		exit(1);
	    }
	}
	argv++;
	argc--;
    }
    if (touchsame && comp_mtime) {
	(void) fprintf(stderr,
	    "%s: options -t and -m are mutually exclusive\n", prog);
	exit(1);
    }

    if (argc != 2) {
	USAGE;
	exit(1);
    }

    (void) strncpy(f1, argv[0], MAXPATHLEN);
    (void) strncpy(f2, argv[1], MAXPATHLEN);

}
/****************************************************************************/
int compdir(d1, d2)
char *d1;
char *d2;
{
    /* Compares two directories.
     * Returns -1 and prints error message if scandir fails or stat fails.
     * Returns 0 on success.
     * prints newer, only as shown at top of program.
     */

#define ISDIR(statstruct) ((statstruct).st_mode & S_IFDIR)

    struct dirent **nml1, **nml2;	/* namelists for directories 1, 2 */
    int n1, n2;				/* number of entries in d1, d2 */
    char f1[MAXPATHLEN], f2[MAXPATHLEN];
    struct stat st1, st2;
    int alphasort();
    int select_dir();
    int cmpcontents();
    void printresult();


    int i, ignore1, ignore2, i1, i2;

    int diff;

    /* Obtain sorted listings of the two directories */
    if ((n1 = scandir(d1, &nml1, select_dir, alphasort)) == -1) {
	(void) fprintf(stderr, "%s: scandir failed on directory %s\n",
			prog, d1);
	return -1;
    } else if ((n2 = scandir(d2, &nml2, select_dir, alphasort)) == -1) {
	(void) fprintf(stderr, "%s: scandir failed on directory %s\n",
			prog, d2);
	return -1;
    }
    

    /* Compare the entries.  i1 goes through d1 list; i2 goes through d2 list.
     * 1. compare i1'th vs i2'th names.  If not same, then file is an "only";
     *    increment the ix'th, print the name, and go to top.
     * 2. Otherwise, test if one is directory but not the other.
     * 3. Otherwise, both are same name. If both are directories, recursively
     *    apply compdir().
     * 4. Otherwise, compare files.
     */

    for (i1 = i2 = 0; i1 < n1 && i2 < n2; i1++, i2++) {

	(void) strcpy(f1, d1);
	(void) strcat(f1, "/");
	(void) strcat(f1, nml1[i1]->d_name);
	(void) strcpy(f2, d2);
	(void) strcat(f2, "/");
	(void) strcat(f2, nml2[i2]->d_name);

    /* fprintf(stderr, "***\t%s %s\n", nml1[i1]->d_name, nml2[i2]->d_name); */
	diff = strcmp(nml1[i1]->d_name, nml2[i2]->d_name);
	if (diff < 0) {
	    /* No such file in d2 */
	    if (bothonlyfile) {
		/* Put the corresponding filename into f2 for printing */
		(void) strcpy(f2, d2);
		(void) strcat(f2, "/");
		(void) strcat(f2, nml1[i1]->d_name);
	    }
	    if (!(nexclude && wildsearch(f1, excludepat, nexclude)))
		printresult(f1, (bothonlyfile ? f2 : ""), 'o');
	    i2--;	/* Decrement i2, so that we use same line again for
			 * next comparison.
			 */
	    continue;
	} else if (diff > 0) {
	    /* No such file in d1 */
	    if (bothonlyfile) {
		/* Put the corresponding filename into f1 for printing */
		(void) strcpy(f1, d1);
		(void) strcat(f1, "/");
		(void) strcat(f1, nml2[i2]->d_name);
	    }
	    if (!(nexclude && wildsearch(f2, excludepat, nexclude)))
		printresult(f2, (bothonlyfile ? f1 : ""), 'o');
	    i1--;	/* Decrement i1, so that we use same line again for
			 * next comparison.
			 */
	    continue;
	}

	/* If on exclude list, skip the file or files */
	ignore1 = (nexclude && wildsearch(f1, excludepat, nexclude));
	ignore2 = (nexclude && wildsearch(f2, excludepat, nexclude));
	if (ignore1 && ignore2)
	    continue;
	else if (ignore1) {
	    i2--;
	    continue;
	} else if (ignore2) {
	    i1--;
	    continue;
	}

	if (stat(f1, &st1) == -1) {
	    (void) Error(1, 0, "stat failed on %s: ", f1);
	    continue;
	}
	if (stat(f2, &st2) == -1) {
	    (void) Error(1, 0, "stat failed on %s: ", f2);
	    continue;
	}
	if (ISDIR(st1) && ISDIR(st2)) {
	    (void) compdir(f1, f2);
	} else if (ISDIR(st1)) {
	    printresult(f1, "", 'D');
	} else if (ISDIR(st2)) {
	    printresult(f2, "", 'D');
	} else if (comp_mtime) {
	    /* Compare times instead of contents */
	    if (st1.st_mtime > st2.st_mtime) {
		/* f1 newer */
		printresult(f1, (bothdifferfile ? f2 : ""), 'd');
	    } else if (st1.st_mtime < st2.st_mtime) {
		/* f2 newer */
		printresult(f2, (bothdifferfile ? f1 : ""), 'd');
	    } else if (samefiles) {
		printresult(f1, (bothsamefile ? f2 : ""), 's');
	    }
	} else {
	    /* Compare contents */
	    if (cmpcontents(f1, &st1, f2, &st2) != 0) {
		/* File contents differ */
		if (st1.st_mtime < st2.st_mtime)
		    printresult(f2, (bothdifferfile ? f1 : ""), 'd');
		else
		    printresult(f1, (bothdifferfile ? f2 : ""), 'd');
	    } else {
		/* File contents are same */
		if (touchsame)	/* adjust newer mtime if they differ */
		    if (st1.st_mtime > st2.st_mtime)
			touchtime(f1, st1.st_atime, st2.st_mtime);
		    else if (st1.st_mtime < st2.st_mtime)
			touchtime(f2, st2.st_atime, st1.st_mtime);
		if (samefiles)
		    printresult(f1, (bothsamefile ? f2 : ""), 's');
	    }
	}
    }
    return 0;
}

/***************************************************************************/
void printresult(f1, f2, type)
char *f1, *f2;
char type;	/* 'D' = f1 is dir (the other file was not);
		 * 'o' = f1 is an only file;
		 * 'd' = files differ; f1 is the newer;
		 * 's' = files are same.
		 * default = same */
{
    printf("%s: %s\t%s\n",
	type == 'D' ? "directory" :
	type == 'd' ? "differ" :
	type == 'o' ? "only" : "same",      f1, f2);
}

/***************************************************************************/
int cmpcontents(f1, st1, f2, st2)
char *f1;
struct stat *st1;
char *f2;
struct stat *st2;
{
    /* compares the contents && sizes of two files, f1 and f2.
     * st1 and st2 are the ptrs to the result of stat()'ing f1 and f2.
     * Failure: returns -1 if they could not be opened/read, in which
     * case an error message is also printed.
     * On success: returns 0 if identical, else 1.
     */
    FILE *fp1, *fp2;

    char b1[BUFSIZ], b2[BUFSIZ];
    int n1, n2;

    int err1, err2;

    /* Check filesize before comparing contents... */
    if (st1->st_size != st2->st_size)
	return 1;

    /* ...then check if f1,f2 are actually links to one file... */
    if (st1->st_dev == st2->st_dev && st1->st_ino == st2->st_ino)
	return 0;

    fp1 = fopen(f1, "r");
    fp2 = fopen(f2, "r");
    if ((char *) fp1 == NULL) {
	(void) Error(1, 0, "couldn't open %s: ", f1);
	return -1;
    }
    if ((char *) fp2 == NULL) {
	(void) Error(1, 0, "couldn't open %s: ", f2);
	return -1;
    }
 
    while ( (n1 = fread(b1, 1, BUFSIZ, fp1)),
	    (n2 = fread(b2, 1, BUFSIZ, fp2)),
	    (n1 && n2) ) {

	if (n1 != n2) /* Hit EOF on a file, but 0 hasn't yet been returned */
	    return 1;

	if (memcmp(b1, b2, n1) != 0) {
	    (void) fclose(fp1);
	    (void) fclose(fp2);
	    return 1;
	}
    }
    /* One or both reads returned 0 */
    err1 = ferror(fp1);
    err2 = ferror(fp2);
    (void) clearerr(fp1);
    (void) clearerr(fp2);
    (void) fclose(fp1);
    (void) fclose(fp2);

    if (n1 == 0 && err1) {
	(void) fprintf(stderr, "%s: read error on %s\n", prog, f1);
	return -1;
    } else if (n2 == 0 && err2) {
	(void) fprintf(stderr, "%s: read error on %s\n", prog, f2);
	return -1;
    } else if (n1 == 0 && n2 == 0) {
	/* Simultaneous EOF on both */
	return 0;
    } else {
	/* EOF on one but not other */
	return 1;
    }
    /* NOTREACHED */
}

/****************************************************************************/
int select_dir(d1)
struct dirent *d1;
{
    /* Select files for scandir.
     * ".", ".." always discarded.
     * ".xxx" is discarded unless allfiles != 0.
     */
    if (strcmp(d1->d_name, ".") == 0)
	return 0;
    else if (strcmp(d1->d_name, "..") == 0)
	return 0;
    else if (!allfiles && d1->d_name[0] == '.')
	return 0;
    else
	return 1;
}

/***************************************************************************/
int wildsearch(string, patlist, npat)
char *string;
char **patlist;
int npat;
{
    int i;
    char *s;

    for (i=0; i < nexclude; i++) {
	/* fprintf(stderr, "\t@%s@\n\t@%s@\n", string, patlist[i]); */
	if ((s = strchr(patlist[i], '/'))) {
	    /* Pattern has a slash; match against entire string */
	    if (wildmat(string, patlist[i]))
		return 1;
	} else {
	    /* Pattern has no slash; match against last component of string */
	    s = strrchr(string, '/');
	    if (wildmat(s ? s+1 : string, patlist[i]))
		return 1;
	}
    }
    return 0;
}

/***************************************************************************/
/*
**  Do shell-style pattern matching for ?, \, [], and * characters.
**  Might not be robust in face of malformed patterns; e.g., "foo[a-"
**  could cause a segmentation violation.
**
**  Written by Rich $alz, mirror!rs, Wed Nov 26 19:03:17 EST 1986.
*/

#define TRUE		1
#define FALSE		0


static int
Star(s, p)
    register char	*s;
    register char	*p;
{
    while (wildmat(s, p) == FALSE)
	if (*++s == '\0')
	    return(FALSE);
    return(TRUE);
}


int
wildmat(s, p)
    register char	*s;
    register char	*p;
{
    register int 	 last;
    register int 	 matched;
    register int 	 reverse;

    for ( ; *p; s++, p++)
	switch (*p) {
	    case '\\':
		/* Literal match with following character; fall through. */
		p++;
	    default:
		if (*s != *p)
		    return(FALSE);
		continue;
	    case '?':
		/* Match anything. */
		if (*s == '\0')
		    return(FALSE);
		continue;
	    case '*':
		/* Trailing star matches everything. */
		return(*++p ? Star(s, p) : TRUE);
	    case '[':
		/* [^....] means inverse character class. */
		if (reverse = p[1] == '^')
		    p++;
		for (last = 0400, matched = FALSE; *++p && *p != ']'; last = *p)
		    /* This next line requires a good C compiler. */
		    if (*p == '-' ? *s <= *++p && *s >= last : *s == *p)
			matched = TRUE;
		if (matched == reverse)
		    return(FALSE);
		continue;
	}

    return(*s == '\0');
}

/**************************************************************************/
int touchtime(f, atime, mtime)
char *f;
time_t atime;
time_t mtime;
{
    /* set mtime for f to be time */

    time_t times[2];

    times[0] = atime;
    times[1] = mtime;
    if(utime(f, times) == -1) {
	(void) Error(1, 0, "couldn't set times on file %s: ", f);
	return -1;
    } else {
	return 0;
    }
    /* NOTREACHED */
}
